from .skyproj import McBrydeSkyproj, LaeaSkyproj
from .utils import get_datafile  # fix this

__all__ = ['DESSkyproj', 'BlissSkyproj', 'MaglitesSkyproj',
           'DecalsSkyproj']


class _Survey:
    """Extension class to add routines for drawing survey outlines.
    """
    def draw_des(self, **kwargs):
        """Draw the DES footprint."""
        return self.draw_des19(**kwargs)

    def draw_des19(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DES 2019 footprint."""
        filename = get_datafile('des-round19-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_des17(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DES 2017 footprint."""
        filename = get_datafile('des-round17-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_decals(self, edgecolor='red', lw=2, **kwargs):
        """Draw the DECaLS footprint."""
        filename = get_datafile('decals-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_maglites(self, edgecolor='blue', lw=2, **kwargs):
        """Draw the MagLiteS footprint."""
        filename = get_datafile('maglites-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    def draw_bliss(self, edgecolor='magenta', lw=2, **kwargs):
        """Draw the BLISS footprint."""
        filename = get_datafile('bliss-poly.txt')
        return self.draw_polygon_file(filename, edgecolor=edgecolor, lw=lw, **kwargs)

    # Override zoom default for survey maps to keep the default fixed.
    def draw_hpxmap(self, hpxmap, nest=False, zoom=False, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        # docstring inherited
        return super().draw_hpxmap(hpxmap,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    def draw_hpxpix(self, nside, pixels, values, nest=False, zoom=False, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        # docstring inherited
        return super().draw_hpxpix(nside,
                                   pixels,
                                   values,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    def draw_hspmap(self, hspmap, zoom=False, xsize=1000, vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        # docstring inherited
        return super().draw_hspmap(hspmap,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)

    def draw_hpxbin(self, lon, lat, C=None, nside=256, nest=False, zoom=False, xsize=1000,
                    vmin=None, vmax=None,
                    rasterized=True, lon_range=None, lat_range=None, **kwargs):
        # docstring inherited
        return super().draw_hpxmap(lon,
                                   lat,
                                   C=C,
                                   nside=nside,
                                   nest=nest,
                                   zoom=zoom,
                                   xsize=xsize,
                                   vmin=vmin,
                                   vmax=vmax,
                                   rasterized=rasterized,
                                   lon_range=lon_range,
                                   lat_range=lat_range,
                                   **kwargs)


class DESSkyproj(McBrydeSkyproj, _Survey):
    # docstring inherited
    def __init__(self, ax=None, lon_0=0, gridlines=True,
                 celestial=True, extent=[90, -50, -74, 10], **kwargs):
        super().__init__(ax=ax, lon_0=lon_0, gridlines=gridlines,
                         celestial=celestial, extent=extent, **kwargs)


class BlissSkyproj(McBrydeSkyproj, _Survey):
    # docstring inherited
    def __init__(self, ax=None, lon_0=100, gridlines=True,
                 celestial=True, extent=[-60, 250, -55, 0], **kwargs):
        super().__init__(ax=ax, lon_0=lon_0, gridlines=gridlines,
                         celestial=celestial, extent=extent, **kwargs)


class MaglitesSkyproj(LaeaSkyproj, _Survey):
    # docstring inherited
    def __init__(self, ax=None, lon_0=0, lat_0=-90, gridlines=True,
                 celestial=True, extent=[-150, 70, -85, -50], **kwargs):
        super().__init__(ax=ax, lon_0=lon_0, lat_0=lat_0,
                         gridlines=gridlines, celestial=celestial, extent=extent, **kwargs)


class DecalsSkyproj(McBrydeSkyproj, _Survey):
    # docstring inherited
    def __init__(self, ax=None, lon_0=105.0, gridlines=True,
                 celestial=True, extent=[180, -180, -30, 40], **kwargs):
        super().__init__(ax=ax, lon_0=lon_0, gridlines=gridlines,
                         celestial=celestial, extent=extent, **kwargs)



import numpy as np
from packaging import version

from mpl_toolkits.axisartist.grid_finder import ExtremeFinderSimple
import mpl_toolkits.axisartist.angle_helper as angle_helper
from mpl_toolkits.axisartist.grid_helper_curvelinear import GridHelperCurveLinear

from .utils import wrap_values

import matplotlib
if version.parse(matplotlib.__version__) >= version.parse('3.5'):
    _celestial_angle_360 = True
else:
    _celestial_angle_360 = False

__all__ = ['WrappedFormatterDMS', 'ExtremeFinderWrapped', 'GridHelperSkyproj']


class WrappedFormatterDMS(angle_helper.FormatterDMS):
    """A tick formatter that handles longitude wrapping.

    Parameters
    ----------
    wrap : `float`
        Angle (degrees) which should be wrapped (subtract 360).
    longitude_ticks : `str`
        Type of longitude ticks, either ``positive`` (0 to 360)
        or ``symmetric`` (-180 to 180).
    """
    def __init__(self, wrap, longitude_ticks):
        self._wrap = wrap
        if longitude_ticks == 'positive':
            self._longitude_ticks = 1
        elif longitude_ticks == 'symmetric':
            self._longitude_ticks = -1
        else:
            raise ValueError("longitude_ticks must be `positive` or `symmetric`.")
        super().__init__()

    def _wrap_values(self, factor, values):
        """Wrap the values according to the wrap angle.

        Parameters
        ----------
        factor : `float`
            Scaling factor for input values
        values : `list`
            List of values to format

        Returns
        -------
        wrapped_values : `np.ndarray`
            Array of wrapped values, scaled by factor.
        """
        _values = np.atleast_1d(values)/factor
        _values = wrap_values(_values, wrap=self._wrap)
        if self._longitude_ticks == 1:
            # Values should all be positive, 0 to 360
            _values %= 360.0
        else:
            # Additional test to set -180 to positive 180
            _values[np.isclose(_values, -180.0)] = 180.0
        return factor*_values

    def __call__(self, direction, factor, values):
        return super().__call__(direction, factor, self._wrap_values(factor, values))


class ExtremeFinderWrapped(ExtremeFinderSimple):
    """
    Find extremes with configurable wrap angle and correct limits.

    Parameters
    ----------
    nx : `int`
        Number of samples in x direction.
    ny : `int`
        Number of samples in y direction.
    wrap_angle : `float`
        Angle at which the 360-degree cycle should be wrapped.
    """
    def __init__(self, nx, ny, wrap_angle):
        self.nx, self.ny = nx, ny
        self._wrap = wrap_angle
        self._eps = 1e-5

    def __call__(self, transform_xy, x1, y1, x2, y2):
        # docstring inherited
        x, y = np.meshgrid(
            np.linspace(x1, x2, self.nx), np.linspace(y1, y2, self.ny))
        lon, lat = transform_xy(np.ravel(x), np.ravel(y))

        with np.errstate(invalid='ignore'):
            lon = wrap_values(lon, wrap=self._wrap)

        lon_min, lon_max = np.nanmin(lon), np.nanmax(lon)
        lat_min, lat_max = np.nanmin(lat), np.nanmax(lat)

        lon_min, lon_max, lat_min, lat_max = \
            self._add_pad(lon_min, lon_max, lat_min, lat_max)

        lat_min = np.clip(lat_min, -90.0 + self._eps, 90.0 - self._eps)
        lat_max = np.clip(lat_max, -90.0 + self._eps, 90.0 - self._eps)

        lon_min = np.clip(lon_min, self._wrap - 360. + self._eps, self._wrap - self._eps)
        lon_max = np.clip(lon_max, self._wrap - 360. + self._eps, self._wrap - self._eps)

        return lon_min, lon_max, lat_min, lat_max


class GridHelperSkyproj(GridHelperCurveLinear):
    """GridHelperCurveLinear with tick overlap protection.

    Parameters
    ----------
    *args : `list`
        Arguments for ``GridHelperCurveLinear``.
    celestial : `bool`, optional
        Plot is celestial, and angles should be 0 to 360.  Otherwise -180 to 180.
    equatorial_labels : `bool`, optional
        Longitude labels are marked on the equator instead of edges.
    delta_cut : `float`, optional
        Gridline step (degrees) to signify a jump around a wrapped edge.
    **kwargs : `dict`, optional
        Additional kwargs for ``GridHelperCurveLinear``.
    """
    def __init__(self, *args, celestial=True, equatorial_labels=False, delta_cut=80.0, **kwargs):
        self._celestial = celestial
        self._equatorial_labels = equatorial_labels
        self._delta_cut = delta_cut

        super().__init__(*args, **kwargs)

    def get_gridlines(self, which="major", axis="both"):
        # docstring inherited
        grid_lines = []
        if axis in ["both", "x"]:
            for gl in self._grid_info["lon"]["lines"]:
                grid_lines.extend(self._cut_grid_line_jumps(gl))
        if axis in ["both", "y"]:
            for gl in self._grid_info["lat"]["lines"]:
                grid_lines.extend(self._cut_grid_line_jumps(gl))
        return grid_lines

    def _cut_grid_line_jumps(self, gl):
        """Check for jumps and cut gridlines into multiple sections.

        Parameters
        ----------
        gl : `list` [`tuple`]
            Input gridlines.  List of tuples of numpy arrays.

        Returns
        -------
        gl_new : `list` [`tuple`]
            New gridlines.  Jumps have been replaced with `np.nan`
            values to ensure lines are not connected around edges.
        """
        dx = gl[0][0][1:] - gl[0][0][: -1]
        dy = gl[0][1][1:] - gl[0][1][: -1]

        split, = (np.hypot(dx, dy) > self._delta_cut).nonzero()

        if split.size == 0:
            return gl

        gl_new = [(np.insert(gl[0][0], split + 1, np.nan),
                   np.insert(gl[0][1], split + 1, np.nan))]

        return gl_new

    def get_tick_iterator(self, nth_coord, axis_side, minor=False):
        # docstring inherited
        try:
            _grid_info = self._grid_info
        except AttributeError:
            _grid_info = self.grid_info

        angle_tangent = dict(left=90, right=90, bottom=0, top=0)[axis_side]
        lon_or_lat = ["lon", "lat"][nth_coord]
        if lon_or_lat == "lon":
            # Need to compute maximum extent in the x direction
            x_min = 1e100
            x_max = -1e100
            for line in _grid_info['lon_lines']:
                x_min = min((x_min, np.min(line[0])))
                x_max = max((x_max, np.max(line[0])))
            delta_x = x_max - x_min
        if not minor:  # major ticks
            tick_locs = _grid_info[lon_or_lat]["tick_locs"][axis_side]
            tick_labels = _grid_info[lon_or_lat]["tick_labels"][axis_side]

            if lon_or_lat == "lon" and self._equatorial_labels:
                min_lat = np.min(_grid_info["lat"]["levels"])
                max_lat = np.max(_grid_info["lat"]["levels"])
                if (min_lat < -89.0 and axis_side == "bottom") \
                   or (max_lat > 89.0 and axis_side == "top"):
                    tick_locs = []
                    tick_labels = []

            prev_xy = None
            for ctr, ((xy, a), l) in enumerate(zip(tick_locs, tick_labels)):
                if self._celestial and _celestial_angle_360:
                    angle_normal = 360.0 - a
                else:
                    angle_normal = a

                if ctr > 0 and lon_or_lat == 'lon':
                    # Check if this is too close to the last label.
                    if abs(xy[0] - prev_xy[0])/delta_x < 0.05:
                        continue
                prev_xy = xy
                yield xy, angle_normal, angle_tangent, l
        else:
            for (xy, a), l in zip(
                    _grid_info[lon_or_lat]["tick_locs"][axis_side],
                    _grid_info[lon_or_lat]["tick_labels"][axis_side]):
                if self._celestial and _celestial_angle_360:
                    angle_normal = 360.0 - a
                else:
                    angle_normal = a
                yield xy, angle_normal, angle_tangent, ""



import numpy as np

from ._skyproj import _Skyproj

__all__ = ['Skyproj', 'McBrydeSkyproj', 'LaeaSkyproj', 'MollweideSkyproj',
           'HammerSkyproj', 'EqualEarthSkyproj', 'GnomonicSkyproj',
           'ObliqueMollweideSkyproj']


class _Stadium:
    """Extension class to create a stadium-shaped projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        edge_offset = 180.0 - 1e-6
        nstep = 1000

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 - edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['left'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 + edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['right'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(90.0 - 1e-6,
                                     90.0 - 1e-6,
                                     nstep))
        proj_boundary_xy['top'] = np.column_stack((x, y))

        x, y = self.proj(np.linspace(self.lon_0 - edge_offset,
                                     self.lon_0 + edge_offset,
                                     nstep),
                         np.linspace(-90.0 + 1e-6,
                                     -90.0 + 1e-6,
                                     nstep))
        proj_boundary_xy['bottom'] = np.column_stack((x, y))

        return proj_boundary_xy


class _Ellipse21:
    """Extension class to create an ellipse-shaped projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        nstep = 1000

        t = np.linspace(-np.pi/2., np.pi/2., nstep)
        x = 2*self.crs.radius*np.sqrt(2)*np.cos(t)
        y = self.crs.radius*np.sqrt(2)*np.sin(t)
        proj_boundary_xy['right'] = np.column_stack((x, y))

        t = np.linspace(np.pi/2., 3*np.pi/2., nstep)
        x = 2*self.crs.radius*np.sqrt(2)*np.cos(t)
        y = self.crs.radius*np.sqrt(2)*np.sin(t)
        proj_boundary_xy['left'] = np.column_stack((x, y))

        proj_boundary_xy['top'] = np.zeros((0, 2))
        proj_boundary_xy['bottom'] = np.zeros((0, 2))

        return proj_boundary_xy


class _Circle:
    """Extension class to create a circular projection boundary.
    """
    def _compute_proj_boundary_xy(self):
        proj_boundary_xy = {}

        nstep = 1000

        t = np.linspace(-np.pi/2., np.pi/2., nstep)
        x = 2*self.crs.radius*np.cos(t)
        y = 2*self.crs.radius*np.sin(t)
        proj_boundary_xy['right'] = np.column_stack((x, y))

        t = np.linspace(np.pi/2., 3*np.pi/2., nstep)
        x = 2*self.crs.radius*np.cos(t)
        y = 2*self.crs.radius*np.sin(t)
        proj_boundary_xy['left'] = np.column_stack((x, y))

        proj_boundary_xy['top'] = np.zeros((0, 2))
        proj_boundary_xy['bottom'] = np.zeros((0, 2))

        return proj_boundary_xy


# The default skyproj is a cylindrical Plate Carree projection.

class Skyproj(_Skyproj, _Stadium):
    # docstring inherited
    # Plate Carree
    def __init__(self, **kwargs):
        super().__init__(projection_name='cyl', **kwargs)


# The following skyprojs include the equal-area projections that are tested
# and known to work.

class McBrydeSkyproj(_Skyproj, _Stadium):
    # docstring inherited
    # McBryde-Thomas Flat Polar Quartic
    def __init__(self, **kwargs):
        super().__init__(projection_name='mbtfpq', **kwargs)


class LaeaSkyproj(_Skyproj, _Circle):
    # docstring inherited
    # Lambert Azimuthal Equal Area
    def __init__(self, **kwargs):
        super().__init__(projection_name='laea', **kwargs)

    @property
    def _full_circle(self):
        return True

    @property
    def _init_extent_xy(self):
        return True

    @property
    def _radial_labels(self):
        return True

    @property
    def _default_xy_labels(self):
        return ("", "")

    @property
    def _full_sky_extent_initial(self):
        lon0 = self.lon_0 - 180.0
        lon1 = self.lon_0 + 180.0
        _lat_0 = self.lat_0
        if _lat_0 == -90.0:
            lat0 = -90.0
            lat1 = 90.0 - 1e-5
        elif _lat_0 == 90.0:
            lat0 = -90.0 + 1e-5
            lat1 = 90.0
        else:
            lat0 = -90.0 + 1e-5
            lat1 = 90.0 - 1e-5

        return [lon0, lon1, lat0, lat1]


class MollweideSkyproj(_Skyproj, _Ellipse21):
    # docstring inherited
    # Mollweide
    def __init__(self, **kwargs):
        super().__init__(projection_name='moll', **kwargs)

    @property
    def _pole_clip(self):
        return 1.0

    @property
    def _equatorial_labels(self):
        return True


class HammerSkyproj(_Skyproj, _Ellipse21):
    # docstring inherited
    # Hammer-Aitoff
    def __init__(self, **kwargs):
        super().__init__(projection_name='hammer', **kwargs)

    @property
    def _equatorial_labels(self):
        return True


class EqualEarthSkyproj(_Skyproj, _Stadium):
    # docstring inherited
    # Equal Earth
    def __init__(self, **kwargs):
        super().__init__(projection_name='eqearth', **kwargs)


class ObliqueMollweideSkyproj(_Skyproj, _Ellipse21):
    """Oblique Mollweide Projection.

    Parameters
    ----------
    lon_0 : `float`, optional
        Central longitude of the underlying Mollweide projection.
    lat_p : `float`, optional
        Latitude of the North Pole of the unrotated coordinate system.
    lon_p : `float`, optional
        Longitude of the North Pole of the unrotated coordinate system.
    **kwargs : `dict`, optional
        Additional kwargs for `skyproj._Skyproj`.
    """
    # Oblique Mollweide
    def __init__(self, **kwargs):
        super().__init__(projection_name='obmoll', **kwargs)

    @property
    def _pole_clip(self):
        return 1.0

    @property
    def _equatorial_labels(self):
        return True

    @property
    def _default_xy_labels(self):
        return ("", "")

    @property
    def _init_extent_xy(self):
        return True


# The Gnomonic (tangent plane) projection is not equal-area and
# is not available for full-sky plots.  It is only for small
# zoomed regions
class GnomonicSkyproj(_Skyproj, _Circle):
    # docstring inherited
    # Gnomonic
    def __init__(self, **kwargs):
        super().__init__(projection_name='gnom', **kwargs)

    @property
    def _full_sky_extent_initial(self):
        lon_0 = self.lon_0
        lat_0 = self.lat_0
        cos_lat = np.cos(np.deg2rad(lat_0))
        return [lon_0 - 0.5/cos_lat, lon_0 + 0.5/cos_lat,
                lat_0 - 0.5, lat_0 + 0.5]



import math
import numpy as np

from pyproj import CRS
from pyproj import Transformer
from pyproj.exceptions import ProjError

from .utils import wrap_values

__all__ = ["SkyProjection", "PlateCarree", "McBrydeThomasFlatPolarQuartic", "Mollweide",
           "Hammer", "EqualEarth", "LambertAzimuthalEqualArea",
           "get_projection", "get_available_projections"]


RADIUS = 1.0


class SkyProjection(CRS):
    def __init__(self, radius=RADIUS, **kwargs):
        self._lon_0_init = kwargs.get('lon_0', 0.0)
        self.proj4_params = {'a': radius,
                             'b': radius}
        self.proj4_params.update(**kwargs)
        super().__init__(self.proj4_params)

    def update_lon_0(self, lon_0):
        """Update projection longitude.

        Parameters
        ----------
        lon_0 : `float`
            Updated lon_0.
        """
        self.proj4_params['lon_0'] = lon_0

        new_proj = CRS(self.proj4_params)
        self._local.crs = new_proj._local.crs

    def reset_lon_0(self):
        """Reset to original lon_0."""
        self.update_lon0(self._lon_0_init)

    def transform_points(self, src_crs, x, y):
        result_shape = tuple(x.shape[i] for i in range(x.ndim)) + (2, )

        x = x.ravel()
        y = y.ravel()

        npts = x.shape[0]

        result = np.zeros([npts, 2], dtype=np.float64)
        if npts:
            if isinstance(src_crs, PlateCarree):
                # We need to wrap to [-180, 180)
                x = wrap_values(x)
            try:
                transformer = Transformer.from_crs(src_crs, self, always_xy=True)
                result[:, 0], result[:, 1] = transformer.transform(x, y, None, errcheck=False)
            except ProjError as err:
                msg = str(err).lower()
                if (
                    "latitude" in msg
                    or "longitude" in msg
                    or "outside of projection domain" in msg
                    or "tolerance condition error" in msg
                ):
                    result[:] = np.nan
                else:
                    raise

            # and set to nans
            result[~np.isfinite(result)] = np.nan

        if len(result_shape) > 2:
            return result.reshape(result_shape)

        return result

    def _as_mpl_transform(self, axes=None):
        from .transforms import SkyTransform

        if axes is None:
            raise ValueError("help need axes")

        return (SkyTransform(self) + axes.transData)

    def _as_mpl_axes(self):
        from .skyaxes import SkyAxes

        return SkyAxes, {'sky_projection': self}


class PlateCarree(SkyProjection):
    def __init__(self, lon_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'eqc',
                        'lon_0': lon_0,
                        'to_meter': math.radians(1)*radius,
                        'vto_meter': 1}

        super().__init__(radius=radius, **proj4_params)


class McBrydeThomasFlatPolarQuartic(SkyProjection):
    def __init__(self, lon_0=0.0, lat_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'mbtfpq',
                        'lon_0': lon_0,
                        'lat_0': lat_0}

        super().__init__(radius=radius, **proj4_params)


class Mollweide(SkyProjection):
    def __init__(self, lon_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'moll',
                        'lon_0': lon_0}

        super().__init__(radius=radius, **proj4_params)


class Hammer(SkyProjection):
    def __init__(self, lon_0=0.0, lat_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'hammer',
                        'lon_0': lon_0,
                        'lat_0': lat_0}

        super().__init__(radius=radius, **proj4_params)


class EqualEarth(SkyProjection):
    def __init__(self, lon_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'eqearth',
                        'lon_0': lon_0}

        super().__init__(radius=radius, **proj4_params)


class LambertAzimuthalEqualArea(SkyProjection):
    def __init__(self, lon_0=0.0, lat_0=0.0, radius=RADIUS):
        proj4_params = {'proj': 'laea',
                        'lon_0': lon_0,
                        'lat_0': lat_0}

        super().__init__(radius=radius, **proj4_params)


_projections = {
    'hammer': ('Hammer', Hammer),
    'mbtfpq': ('McBryde-Thomas Flat Polar Quartic', McBrydeThomasFlatPolarQuartic),
    'cyl': ('Plate Carree', PlateCarree),
    'eqearth': ('Equal Earth', EqualEarth),
    'laea': ('Lambert Azimuthal Equal Area', LambertAzimuthalEqualArea),
    'moll': ('Mollweide', Mollweide)
}


def get_projection(name, **kwargs):
    """Return a skyproj projection.

    For list of projections available, use skyproj.get_available_projections().

    Parameters
    ----------
    name : `str`
        Skyproj name of projection.
    **kwargs :
        Additional kwargs appropriate for given projection.

    Returns
    -------
    proj : `skyproj.SkyProjection`
    """
    # Is this a listed projection?
    if name not in _projections:
        raise ValueError(f'{name} projection name is not recognized.  See get_available_projections()')

    import IPython
    IPython.embed()

    descr, projclass = _projections[name]

    return projclass(**kwargs)


def get_available_projections():
    """Return dict of available projections.

    Returns
    -------
    available_projections: `dict`
        Available projections.  Key is cartosky name, value is brief description.
    """
    available_projections = {}
    for name, (descr, projclass) in _projections.items():
        available_projections[name] = descr

    return available_projections



import os
import pytest
import numpy as np

import matplotlib
matplotlib.use("Agg")

from matplotlib.testing.compare import compare_images, ImageComparisonFailure  # noqa: E402
import matplotlib.pyplot as plt  # noqa: E402

import skyproj  # noqa: E402


ROOT = os.path.abspath(os.path.dirname(__file__))


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj,
                                     skyproj.HammerSkyproj,
                                     skyproj.EqualEarthSkyproj])
@pytest.mark.parametrize("lon_0", [0.0, -100.0, 100.0, 180.0])
def test_skyproj_basic(tmp_path, skyproj, lon_0):
    """Test full sky maps."""
    plt.rcParams.update(plt.rcParamsDefault)

    # Full image
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, lon_0=lon_0)
    fname = f'{sp.projection_name}_full_{lon_0}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 10.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("skyproj", [skyproj.Skyproj,
                                     skyproj.McBrydeSkyproj,
                                     skyproj.MollweideSkyproj,
                                     skyproj.HammerSkyproj,
                                     skyproj.EqualEarthSkyproj,
                                     skyproj.LaeaSkyproj])
def test_skyproj_zoom(tmp_path, skyproj):
    plt.rcParams.update(plt.rcParamsDefault)

    # Simple zoom
    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj(ax=ax, extent=[0, 50, 0, 50])
    fname = f'{sp.projection_name}_zoom.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 10.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lonlat", [(0.0, 0.0),
                                    (120.0, -75.0),
                                    (-120.0, 75.0)])
def test_skyproj_gnom(tmp_path, lonlat):
    """Test gnomonic zooms."""
    plt.rcParams.update(plt.rcParamsDefault)

    lon_0, lat_0 = lonlat

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.GnomonicSkyproj(ax=ax, lon_0=lon_0, lat_0=lat_0)
    # draw a square square, make sure it looks square
    delta_lat = 0.1
    delta_lon = delta_lat/np.cos(np.deg2rad(lat_0))
    sp.draw_polygon(
        [lon_0 - delta_lon, lon_0 + delta_lon, lon_0 + delta_lon, lon_0 - delta_lon],
        [lat_0 - delta_lat, lat_0 - delta_lat, lat_0 + delta_lat, lat_0 + delta_lat]
    )
    fname = f'gnom_{lon_0}_{lat_0}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 10.0)
    if err:
        raise ImageComparisonFailure(err)


@pytest.mark.parametrize("lonlatplonp", [(0.0, 45.0, -90.0),
                                         (100.0, 80.0, 0.0)])
def test_skyproj_obmoll(tmp_path, lonlatplonp):
    """Test Oblique Mollweide."""
    plt.rcParams.update(plt.rcParamsDefault)

    lon_0, lat_p, lon_p = lonlatplonp

    fig = plt.figure(1, figsize=(8, 5))
    fig.clf()
    ax = fig.add_subplot(111)
    sp = skyproj.ObliqueMollweideSkyproj(ax=ax, lon_0=lon_0, lat_p=lat_p, lon_p=lon_p)
    fname = f'{sp.projection_name}_{lon_0}_{lat_p}_{lon_p}.png'
    fig.savefig(tmp_path / fname)
    err = compare_images(os.path.join(ROOT, 'data', fname), tmp_path / fname, 10.0)
    if err:
        raise ImageComparisonFailure(err)
